import os
import json

### A global helper functions to use across the project

def parse_list(input_value, separator=',') -> list[str]:
    try:
        if isinstance(input_value, str):
            return [item.strip() for item in input_value.split(separator) if item.strip()]
        elif isinstance(input_value, list):
            return [item.strip() for item in input_value if item.strip()]
        return []
    except Exception as e:
        raise ValueError(f"Failed to parse list value: {e}")


def to_boolean(value) -> bool:
    try:
        if isinstance(value, str):
            return value.lower() in ['true', '1']
        return bool(value)
    except Exception as e:
        raise ValueError(f"Failed to parse boolean value: {e}")


def to_integer(value) -> int:
    try:
        return int(value)
    except (ValueError, TypeError) as e:
        raise ValueError(f"Failed to parse integer value: {e}")


def is_llm_type(llm_type, enum_value):
    """Check if llm_type equals the enum value or its name, case-insensitive."""
    if llm_type == enum_value:
        return True
    
    if isinstance(llm_type, str):
        llm_type_lower = llm_type.lower()
        enum_name_lower = enum_value.name.lower() if enum_value.name else ""
        enum_value_lower = enum_value.value.lower() if isinstance(enum_value.value, str) else ""

        return (
            llm_type_lower == enum_name_lower or
            llm_type_lower == enum_value_lower or
            llm_type_lower.startswith(enum_name_lower) or # Usecase for snowflake which its type is the provider name + the model name
            llm_type_lower.startswith(enum_value_lower) or
            llm_type_lower in enum_value_lower # Check if the enum value includes the llm type - when providing partial name
        )

    return False
  

def is_support_temperature(llm_type: str, llm_model: str) -> bool:
    """
    Check if the LLM model supports temperature setting.
    """
    current_dir = os.path.dirname(os.path.abspath(__file__))
    json_file_path = os.path.join(current_dir, 'temperature_supported_models.json')

    try:
        with open(json_file_path, 'r') as f:
            temperature_supported_models = json.load(f)
        
        # Check if llm_type exists and llm_model is in its list
        if llm_type in temperature_supported_models:
            return llm_model in temperature_supported_models[llm_type]
        
        return False
    except (FileNotFoundError, json.JSONDecodeError, KeyError):
        return False


from typing import Optional, Union, Dict, Any
from langchain.chains.base import Chain
from langchain.llms.base import LLM

from ..utils.general import parse_list, to_integer, to_boolean
from ..utils.timbr_llm_utils import generate_sql
from ..utils.timbr_utils import validate_sql


class ValidateTimbrSqlChain(Chain):
    """
    LangChain chain for validating SQL queries against Timbr knowledge graph schemas.
    
    This chain validates SQL queries to ensure they are syntactically correct and 
    compatible with the target Timbr ontology/knowledge graph structure. It uses an LLM
    for validation and connects to Timbr via URL and token.
    """
    
    def __init__(
        self,
        llm: LLM,
        url: str,
        token: str,
        ontology: str,
        schema: Optional[str] = 'dtimbr',
        concept: Optional[str] = None,
        retries: Optional[int] = 3,
        concepts_list: Optional[Union[list[str], str]] = None,
        views_list: Optional[Union[list[str], str]] = None,
        include_logic_concepts: Optional[bool] = False,
        include_tags: Optional[Union[list[str], str]] = None,
        exclude_properties: Optional[Union[list[str], str]] = ['entity_id', 'entity_type', 'entity_label'],
        max_limit: Optional[int] = 500,
        note: Optional[str] = '',
        db_is_case_sensitive: Optional[bool] = False,
        graph_depth: Optional[int] = 1,
        verify_ssl: Optional[bool] = True,
        is_jwt: Optional[bool] = False,
        jwt_tenant_id: Optional[str] = None,
        conn_params: Optional[dict] = None,
        debug: Optional[bool] = False,
        **kwargs,
    ):
        """
        :param llm: An LLM instance or a function that takes a prompt string and returns the LLM’s response
        :param url: Timbr server url
        :param token: Timbr password or token value
        :param ontology: The name of the ontology/knowledge graph
        :param schema: The name of the schema to query
        :param concept: The name of the concept to query
        :param retries: The maximum number of retries to attempt
        :param concepts_list: Optional specific concept options to query
        :param views_list: Optional specific view options to query
        :param include_logic_concepts: Optional boolean to include logic concepts (concepts without unique properties which only inherits from an upper level concept with filter logic) in the query. 
        :param include_tags: Optional specific concepts & properties tag options to use in the query (Disabled by default. Use '*' to enable all tags or a string represents a list of tags divided by commas (e.g. 'tag1,tag2')
        :param exclude_properties: Optional specific properties to exclude from the query (entity_id, entity_type & entity_label by default).
        :param max_limit: Maximum number of rows to query
        :param note: Optional additional note to extend our llm prompt
        :param db_is_case_sensitive: Whether the database is case sensitive (default is False).
        :param graph_depth: Maximum number of relationship hops to traverse from the source concept during schema exploration (default is 1).
        :param verify_ssl: Whether to verify SSL certificates (default is True).
        :param is_jwt: Whether to use JWT authentication (default is False).
        :param jwt_tenant_id: JWT tenant ID for multi-tenant environments (required when is_jwt=True).
        :param conn_params: Extra Timbr connection parameters sent with every request (e.g., 'x-api-impersonate-user').
        :param kwargs: Additional arguments to pass to the base
        
        ## Example
        ```
        validate_timbr_sql_chain = ValidateTimbrSqlChain(
            url=<url>,
            token=<token>,
            llm=<llm or timbr_llm_wrapper instance>,
            ontology=<ontology_name>,
            schema=<schema_name>,
            concept=<concept_name>,
            retries=<retries_number>,
            concepts_list=<concepts>,
            views_list=<views>,
            include_tags=<tags>,
            note=<note>,
        )

        return validate_timbr_sql_chain.invoke({ "prompt": question, "sql": <latest_query_to_validate> }).get("sql", [])
        ```
        """
        super().__init__(**kwargs)
        self._llm = llm
        self._url = url
        self._token = token
        self._ontology = ontology
        self._schema = schema
        self._concept = concept
        self._retries = retries
        self._concepts_list = parse_list(concepts_list)
        self._views_list = parse_list(views_list)
        self._include_logic_concepts = to_boolean(include_logic_concepts)
        self._include_tags = parse_list(include_tags)
        self._exclude_properties = parse_list(exclude_properties)
        self._max_limit = to_integer(max_limit)
        self._note = note
        self._db_is_case_sensitive = to_boolean(db_is_case_sensitive)
        self._graph_depth = to_integer(graph_depth)
        self._verify_ssl = to_boolean(verify_ssl)
        self._is_jwt = to_boolean(is_jwt)
        self._jwt_tenant_id = jwt_tenant_id
        self._debug = to_boolean(debug)
        self._conn_params = conn_params or {}


    @property
    def usage_metadata_key(self) -> str:
        return "validate_sql_usage_metadata"


    @property
    def input_keys(self) -> list:
        return ["prompt", "sql"]


    @property
    def output_keys(self) -> list:
        return [
            "sql",
            "schema",
            "concept",
            "is_sql_valid",
            "error",
            self.usage_metadata_key,
        ]


    def _get_conn_params(self) -> dict:
        return {
            "url": self._url,
            "token": self._token,
            "ontology": self._ontology,
            "verify_ssl": self._verify_ssl,
            "is_jwt": self._is_jwt,
            "jwt_tenant_id": self._jwt_tenant_id,
            **self._conn_params,
        }


    def _call(self, inputs: Dict[str, Any], run_manager=None) -> Dict[str, Any]:
        usage_metadata = {}
        sql = inputs["sql"]
        prompt = inputs["prompt"]
        schema = self._schema
        concept = self._concept

        is_sql_valid, error = validate_sql(sql, self._get_conn_params())
        if not is_sql_valid:
            prompt_extension = self._note + '\n' if self._note else ""
            generate_res = generate_sql(
                question=prompt,
                llm=self._llm,
                conn_params=self._get_conn_params(),
                schema=schema,
                concept=concept,
                concepts_list=self._concepts_list,
                views_list=self._views_list,
                include_logic_concepts=self._include_logic_concepts,
                include_tags=self._include_tags,
                exclude_properties=self._exclude_properties,
                should_validate_sql=True,
                retries=self._retries,
                max_limit=self._max_limit,
                note=f"{prompt_extension}The original SQL query (`{sql}`) was invalid with this error from query {error}. Please take this in consideration while generating the query.",
                db_is_case_sensitive=self._db_is_case_sensitive,
                graph_depth=self._graph_depth,
                debug=self._debug,
            )
            sql = generate_res.get("sql", "")
            schema = generate_res.get("schema", self._schema)
            concept = generate_res.get("concept", self._concept)
            usage_metadata.update(generate_res.get("usage_metadata", {}))
            is_sql_valid = generate_res.get("is_sql_valid")
            error = generate_res.get("error")

        return {
            "sql": sql,
            "schema": schema,
            "concept": concept,
            "is_sql_valid": is_sql_valid,
            "error": error,
            self.usage_metadata_key: usage_metadata,
        }



from enum import Enum
from langchain.llms.base import LLM
from pydantic import Field

from .timbr_llm_wrapper import TimbrLlmWrapper
from ..utils.general import is_llm_type, is_support_temperature
from ..config import llm_temperature

class LlmTypes(Enum):
  OpenAI = 'openai-chat'
  Anthropic = 'anthropic-chat'
  Google = 'chat-google-generative-ai'
  AzureOpenAI = 'azure-openai-chat'
  Snowflake = 'snowflake-cortex'
  Databricks = 'chat-databricks'
  Timbr = 'timbr'


class LlmWrapper(LLM):
  """
  LlmWrapper is a unified interface for connecting to various Large Language Model (LLM) providers
  (OpenAI, Anthropic, Google, Azure OpenAI, Snowflake Cortex, Databricks, etc.) using LangChain. It abstracts
  the initialization and connection logic for each provider, allowing you to switch between them
  with a consistent API.
  """
  client: LLM = Field(default=None, exclude=True)

  def __init__(
      self,
      llm_type: str,
      api_key: str,
      model: str = None,
      **llm_params,
  ):
      """
      :param llm_type (str): The type of LLM provider (e.g., 'openai-chat', 'anthropic-chat').
      :param api_key (str): The API key for authenticating with the LLM provider.
      :param model (str): The model name or deployment to use. Defaults to provider-specific values (Optional).
      :param **llm_params: Additional parameters for the LLM (e.g., temperature, endpoint, etc.).
      """
      super().__init__()
      self.client = self._connect_to_llm(
        llm_type,
        api_key,
        model,
        **llm_params,
      )


  @property
  def _llm_type(self):
    return self.client._llm_type


  def _add_temperature(self, llm_type, llm_model, **llm_params):
    """
    Add temperature to the LLM parameters if the LLM model supports it.
    """
    if "temperature" not in llm_params:
      if llm_temperature is not None and is_support_temperature(llm_type, llm_model):
        llm_params["temperature"] = llm_temperature
    return llm_params

  
  def _connect_to_llm(self, llm_type, api_key, model, **llm_params):
    if is_llm_type(llm_type, LlmTypes.OpenAI):
      from langchain_openai import ChatOpenAI as OpenAI
      llm_model = model or "gpt-4o-2024-11-20"
      params = self._add_temperature(LlmTypes.OpenAI.name, llm_model, **llm_params)
      return OpenAI(
        openai_api_key=api_key,
        model_name=llm_model,
        **params,
      )
    elif is_llm_type(llm_type, LlmTypes.Anthropic):
      from langchain_anthropic import ChatAnthropic as Claude
      llm_model = model or "claude-3-5-sonnet-20241022"
      params = self._add_temperature(LlmTypes.Anthropic.name, llm_model, **llm_params)
      return Claude(
        anthropic_api_key=api_key,
        model=llm_model,
        **params,
      )
    elif is_llm_type(llm_type, LlmTypes.Google):
      from langchain_google_genai import ChatGoogleGenerativeAI
      llm_model = model or "gemini-2.0-flash-exp"
      params = self._add_temperature(LlmTypes.Google.name, llm_model, **llm_params)
      return ChatGoogleGenerativeAI(
        google_api_key=api_key,
        model=llm_model,
        **params,
      )
    elif is_llm_type(llm_type, LlmTypes.Timbr):
      return TimbrLlmWrapper(
        api_key=api_key,
        **params,
      )
    elif is_llm_type(llm_type, LlmTypes.Snowflake):
      from langchain_community.chat_models import ChatSnowflakeCortex
      llm_model = model or "openai-gpt-4.1"
      params = self._add_temperature(LlmTypes.Snowflake.name, llm_model, **llm_params)
      snowflake_password = params.pop('snowflake_api_key', params.pop('snowflake_password', api_key))
      
      return ChatSnowflakeCortex(
        model=llm_model,
        snowflake_password=snowflake_password,
        **params,
      )
    elif is_llm_type(llm_type, LlmTypes.AzureOpenAI):
      from langchain_openai import AzureChatOpenAI
      azure_endpoint = params.pop('azure_endpoint', None)
      azure_api_version = params.pop('azure_openai_api_version', None)
      llm_model = model or "gpt-4o-2024-11-20"
      params = self._add_temperature(LlmTypes.AzureOpenAI.name, llm_model, **llm_params)
      return AzureChatOpenAI(
        openai_api_key=api_key,
        azure_deployment=llm_model,
        azure_endpoint=azure_endpoint,
        openai_api_version=azure_api_version,
        **params,
      )
    elif is_llm_type(llm_type, LlmTypes.Databricks):
      from databricks.sdk import WorkspaceClient
      from databricks_langchain import ChatDatabricks
      llm_model = model or "databricks-claude-sonnet-4"
      params = self._add_temperature(LlmTypes.Databricks.name, llm_model, **llm_params)

      host = params.pop('databricks_host', params.pop('host', None))
      w = WorkspaceClient(host=host, token=api_key)
      return ChatDatabricks(
        endpoint=llm_model,
        workspace_client=w,  # Using authenticated client
        **params,
      )
    else:
      raise ValueError(f"Unsupported LLM type: {llm_type}")


  def get_model_list(self) -> list[str]:
    """Return the list of available models for the LLM."""
    models = []
    try:
      if is_llm_type(self._llm_type, LlmTypes.OpenAI):
        from openai import OpenAI
        client = OpenAI(api_key=self.client.openai_api_key._secret_value)
        models = [model.id for model in client.models.list()]
      elif is_llm_type(self._llm_type, LlmTypes.Anthropic):
        import anthropic
        client = anthropic.Anthropic(api_key=self.client.anthropic_api_key._secret_value)
        models = [model.id for model in client.models.list()]
      elif is_llm_type(self._llm_type, LlmTypes.Google):
        import google.generativeai as genai
        genai.configure(api_key=self.client.google_api_key._secret_value)
        models = [m.name.replace('models/', '') for m in genai.list_models()]
      elif is_llm_type(self._llm_type, LlmTypes.AzureOpenAI):
        from openai import AzureOpenAI
        # Get Azure-specific attributes from the client
        azure_endpoint = getattr(self.client, 'azure_endpoint', None)
        api_version = getattr(self.client, 'openai_api_version', None)
        api_key = self.client.openai_api_key._secret_value
        
        if azure_endpoint and api_version and api_key:
          client = AzureOpenAI(
            api_key=api_key,
            azure_endpoint=azure_endpoint,
            api_version=api_version
          )
          # For Azure, get the deployments instead of models
          try:
            models = [model.id for model in client.models.list()]
          except:
            # If listing models fails, provide some common deployment names
            models = ["gpt-4o", "Other (Custom)"]
      elif is_llm_type(self._llm_type, LlmTypes.Snowflake):
        # Snowflake Cortex available models
        models = [
          "openai-gpt-4.1",
          "mistral-large2",
          "llama3.1-70b",
          "llama3.1-405b"
        ]
      elif is_llm_type(self._llm_type, LlmTypes.Databricks):
        w = self.client.workspace_client
        models = [ep.name for ep in w.serving_endpoints.list()]
        
      # elif self._is_llm_type(self._llm_type, LlmTypes.Timbr):
        
    except Exception as e:
      models = []
    
    return sorted(models)


  def _call(self, prompt, **kwargs):
    return self.client(prompt, **kwargs)


  def __call__(self, prompt, **kwargs):
        """
        Override the default __call__ method to handle input preprocessing.
        I used this in order to override prompt input validation made by pydantic
        and allow sending list of AiMessages instead of string only
        """
        return self._call(prompt, **kwargs)
  

  def query(self, prompt, **kwargs):
    return self._call(prompt, **kwargs)

